(* 
    OCaml XenStore Daemon.
    Copyright (C) 2008 Patrick Colp University of British Columbia

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)


(* TYPES *)

(* XenStore node type *)
type xs_node = {
  name : string;
  contents : xs_node_data ref;
  
  (* XXX: HUGE HACK for broken tools; allows for data and children *)
  contents2 : xs_node_data option ref;
  
  (* FIXME: Permissions *)
  permissions : xs_node_permission list ref;
}

(* XenStore node data type *)
and xs_node_data = Value of string | Children of xs_node list

(* XenStore node permission type *)
and xs_node_permission = {
  perm_dom_id : int ref;
  perms : xs_perm_type list ref;
}

(* XenStore permission type *)
and xs_perm_type =
  | XS_PERM_NONE
  | XS_PERM_READ
  | XS_PERM_WRITE
  (* Internal use *)
  | XS_PERM_ENOENT_OK
  | XS_PERM_OWNER;;

let xs_perm_type_to_int perm =
  match perm with
  | XS_PERM_NONE -> 0
  | XS_PERM_READ -> 1
  | XS_PERM_WRITE -> 2
  | XS_PERM_ENOENT_OK -> 4
  | XS_PERM_OWNER -> 8;;

let int_to_xs_perm_type i =
  match i with
  | 0 -> XS_PERM_NONE
  | 1 -> XS_PERM_READ
  | 2 -> XS_PERM_WRITE
  | 4 -> XS_PERM_ENOENT_OK
  | 8 -> XS_PERM_OWNER
  | _ -> XS_PERM_NONE;;

(* Changed node type (TRANSACTION) *)
type changed_node = {
  node_name : string;
  recurse : bool;
};;

(* Changed domain type (TRANSACTION) *)
type changed_domain = {
  changed_dom_id : int;
  changed_num_entries : int ref;
};;

(* XenStore message types *)
type xsd_sockmsg_type =
  | XS_DEBUG
  | XS_DIRECTORY
  | XS_READ
  | XS_GET_PERMS
  | XS_WATCH
  | XS_UNWATCH
  | XS_TRANSACTION_START
  | XS_TRANSACTION_END
  | XS_INTRODUCE
  | XS_RELEASE
  | XS_GET_DOMAIN_PATH
  | XS_WRITE
  | XS_MKDIR
  | XS_RM
  | XS_SET_PERMS
  | XS_WATCH_EVENT
  | XS_ERROR
  | XS_IS_DOMAIN_INTRODUCED
  | XS_RESUME
  | XS_SET_TARGET
  | XS_UNKNOWN;;

let sockmsg_type_to_int msg_type =
  match msg_type with
  | XS_DEBUG -> 0L
  | XS_DIRECTORY -> 1L
  | XS_READ -> 2L
  | XS_GET_PERMS -> 3L
  | XS_WATCH -> 4L
  | XS_UNWATCH -> 5L
  | XS_TRANSACTION_START -> 6L
  | XS_TRANSACTION_END -> 7L
  | XS_INTRODUCE -> 8L
  | XS_RELEASE -> 9L
  | XS_GET_DOMAIN_PATH -> 10L
  | XS_WRITE -> 11L
  | XS_MKDIR -> 12L
  | XS_RM -> 13L
  | XS_SET_PERMS -> 14L
  | XS_WATCH_EVENT -> 15L
  | XS_ERROR -> 16L
  | XS_IS_DOMAIN_INTRODUCED -> 17L
  | XS_RESUME -> 18L
  | XS_SET_TARGET -> 19L
  | XS_UNKNOWN -> (- 1L);;

let int_to_sockmsg_type msg_type =
  match msg_type with
  | 0L -> XS_DEBUG
  | 1L -> XS_DIRECTORY
  | 2L -> XS_READ
  | 3L -> XS_GET_PERMS
  | 4L -> XS_WATCH
  | 5L -> XS_UNWATCH
  | 6L -> XS_TRANSACTION_START
  | 7L -> XS_TRANSACTION_END
  | 8L -> XS_INTRODUCE
  | 9L -> XS_RELEASE
  | 10L -> XS_GET_DOMAIN_PATH
  | 11L -> XS_WRITE
  | 12L -> XS_MKDIR
  | 13L -> XS_RM
  | 14L -> XS_SET_PERMS
  | 15L -> XS_WATCH_EVENT
  | 16L -> XS_ERROR
  | 17L -> XS_IS_DOMAIN_INTRODUCED
  | 18L -> XS_RESUME
  | 19L -> XS_SET_TARGET
  | _ -> XS_UNKNOWN;;

(* XenStore socket message header type *)
type xsd_sockmsg = {
  msg_type : xsd_sockmsg_type ref; (* XS_??? -- uint32 *)
  req_id : int64 ref; (* Request identifier, echoed in daemon's response -- uint32 *)
  tx_id : int64 ref; (* Transaction id (0 if not related to a transaction) -- uint32 *)
  len : int64 ref; (* Length of data following this -- uint32 *)
  buff : string; (* Buffer to read/write header from *)
}

(* Error types *)
type xsd_errors = {
  errnum : int; (* XXX: int32? *)
  errstring : string;
}

let get_xsd_error_string error =
  match error with
  | Unix.EINVAL -> "EINVAL"
  | Unix.EACCES -> "EACCES"
  | Unix.EEXIST -> "EEXIST"
  | Unix.EISDIR -> "EISDIR"
  | Unix.ENOENT -> "ENOENT"
  | Unix.ENOMEM -> "ENOMEM"
  | Unix.ENOSPC -> "ENOSPC"
  | Unix.EIO -> "EIO"
  | Unix.ENOTEMPTY -> "ENOTEMPTY"
  | Unix.ENOSYS -> "ENOSYS"
  | Unix.EROFS -> "EROFS"
  | Unix.EBUSY -> "EBUSY"
  | Unix.EAGAIN -> "EAGAIN"
  | Unix.EISCONN -> "EISCONN"
  | _ -> "EINVAL";;

(* Buffered data type *)
type buffered_data = {
  (* Are we still doing the header? *)
  in_header : bool ref;
  
  (* How far are we? *)
  used : int ref; (* uint *)
  
  header : xsd_sockmsg ref;
  
  (* The actual data *)
  buffer : string ref;
}

(* Domain type *)
type domain = {
  (* The id of this domain *)
  dom_id : int ref;
  
  (* Event channel port *)
  port : Unix.file_descr ref; (* uint32 *)
  
  (* The remote end of the event channel, used only to validate repeated   *)
  (* domain introductions                                                  *)
  remote_port : Unix.file_descr ref; (* uint32 *)
  
  (* The mfn associated with the event channel, used only to validate      *)
  (* repeated domain introductions                                         *)
  mfn : int ref; (* unsigned long *)
  
  (* Domain path in store *) path : string ref;
  
  (* Shared page *)
  interface : (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t option ref;
  
  (* The connection associated with this *)
  conn : connection ref;
  
  (* Have we noticed that this domain is shutdown? *)
  shutdown : bool ref;
  
  (* Number of entries from this domain in the store *)
  num_entries : int ref;
  
  (* Number of watches for this domain *)
  num_watches : int ref;
}

and xenstore_domain_interface_foo = {
  place_holder : (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t;
}

(* XenStore domain interface type *)
and xenstore_domain_interface = {
  (* Requests to xenstore daemon *)
  req : string;
  
  (* Replies and async watch events *)
  rsp : string;
  
  req_cons : int64 ref; (* uint32_t *)
  req_prod : int64 ref; (* uint32_t *)
  
  rsp_cons : int64 ref; (* uint32_t *)
  rsp_prod : int64 ref; (* uint32_t *)
}

(* Transaction type *)
and transaction = {
  (* Connection-local identifier for this transaction. *)
  trans_id : int64 ref; (* uint32_t *)
  
  (* Generation when transaction started *)
  generation : int64 ref; (* unsigned int *)
  
  (* The store to work on *)
  store : xs_node ref;
  
  (* List of changed nodes *)
  changes : changed_node list ref;
  
  (* List of changed domains - to record the changed domain entry number *)
  changed_domains : changed_domain list ref;
}

(* Connection type *)
and connection = {
  (* The file descriptor we came in on *)
  fd : Unix.file_descr option ref;
  
  (* Who am I? 0 for socket connections *)
  id : int ref; (* uint *)
  
  (* Is this a read-only connection? *)
  can_write : bool ref;
  
  (* Buffered incoming data *)
  in_buff : buffered_data ref; (* list? Maybe not ref? *)
  
  (* Buffered output data struct list_head out_list; *)
  out_list : buffered_data list ref;
  
  (* Transaction context for current request (NULL if none) *)
  transaction : transaction option ref;
  
  (* List of in-progress transactions *)
  transaction_list : transaction list ref;
  next_transaction_id : int64 ref; (* uint32 *)
  transaction_started : int64 ref; (* uint32 *)
  
  (* The domain I'm associated with, if any *)
  domain : domain option ref;
  
  (* My watches *)
  watches : watch list ref;
  
  (* Methods for communicating over this connection: write can be NULL *)
  write : (connection -> string -> int -> int -> int) ref;
  read : (connection -> string -> int -> int -> int) ref;
}

and watch = {
  (* Current outstanding events applying to this watch *)
  events : string list ref;
  
  (* Is this relative to connnection's implicit path? *)
  relative_path : string option;
  
  token : string;
  node : string;
}

(* VARIABLES *)

(* # chars in a path name including nul *)
let path_max = 4096;;
let xenstore_payload_max = 4096;;
let xenstore_abs_path_max = 3072;;
let xenstore_rel_path_max = 2048;;

(* dom_id_self is used in certain contexts to refer to oneself *)
let dom_id_self = 0x7FF0;;

(* Usage message header *)
let usage = "Usage:\n"
  ^ "  xenstored <options>\n"
  ^ "\n"
  ^ "where options may include:\n";;

(* Trace file *)
let trace_file = ref "";;
let (trace_fd : Unix.file_descr option ref) = ref (None :> Unix.file_descr option);;

(* Syslog *)
let syslog = ref None;;

(* Quota variables *)
let quota_num_entry_per_domain = ref 1000;;
let quota_max_entry_size = ref 2048;;
let quota_num_watches_per_domain = ref 128;;
let quota_max_transaction = ref 10;;

(* Global flags *)
let recovery = ref true;;
let remove_local = ref true;;
let verbose = ref false;;

(* Size of xsd_sockmsg in bytes *)
let header_size = 16;;

(* List of connections *)
let (connections : connection list ref) = ref [];;

(* List of domains *)
let (domains : domain list ref) = ref [];;

(* XenStore ring size *)
let xenstore_ring_size = 1024;;

(* Store *)
let (store : xs_node ref) = ref { name = "/"; contents = ref (Children []); contents2 = ref None; permissions = ref [ { perm_dom_id = ref 0; perms = ref [ XS_PERM_NONE ] } ] };;

(* Transaction stuff *)
let generation = ref Int64.zero;;

(* FUNCTIONS *)

(* UTILITY FUNCTIONS *)

(* Convert a file descriptor to an int *)
let int_of_file_descr fd = (Obj.magic (fd: Unix.file_descr) : int);;
(* let int_of_file_descr (fd : Unix.file_descr) : int = let o = Obj.repr   *)
(* fd in if (Obj.is_int o) then ( Obj.obj o; ) else ( failwith "Woops";    *)
(* );;                                                                     *)

(* Convert an int to a file descriptor *)
external file_descr_of_int : int -> Unix.file_descr = "%identity"
(* let file_descr_of_int fd = assert (fd >= 0); assert (Obj.is_int         *)
(* (Obj.repr Unix.stdin)); ((Obj.magic fd) : Unix.file_descr);;            *)

(* Extract the first element from a tuple *)
let tuple_first (a, b, c) =
  a;;

(* Extract the second element from a tuple *)
let tuple_second (a, b, c) =
  b;;

(* Remove an item from a list *)
let remove_from_list (lst : 'a list) n =
  let (new_lst : 'a list ref) = ref [] in
  
  let rec loop i = (
      let j = ((List.length lst) - 1 - i) in
      
      if (j >= 0) then (
        if (not (j = n)) then (
          new_lst := (List.nth lst j) :: !new_lst;
        );
        
        loop (succ i);
      );
    ) in
  
  loop 0;
  
  !new_lst;;

(* Strip the trailing null byte off a string *)
let strip_null str =
  if ((String.get str ((String.length str) - 1)) = (char_of_int 0))then (
    String.sub str 0 (String.length str - 1);
  ) else (
    str;
  );;

(* Null terminate a string *)
let null_terminate str =
  str ^ (String.make 1 (char_of_int 0));;

(* Return string representation of a socket message *)
let sockmsg_string msg_type =
  match msg_type with
  | XS_DEBUG -> "DEBUG"
  | XS_DIRECTORY -> "DIRECTORY"
  | XS_READ -> "READ"
  | XS_GET_PERMS -> "GET_PERMS"
  | XS_WATCH -> "WATCH"
  | XS_UNWATCH -> "UNWATCH"
  | XS_TRANSACTION_START -> "TRANSACTION_START"
  | XS_TRANSACTION_END -> "TRANSACTION_END"
  | XS_INTRODUCE -> "INTRODUCE"
  | XS_RELEASE -> "RELEASE"
  | XS_GET_DOMAIN_PATH -> "GET_DOMAIN_PATH"
  | XS_WRITE -> "WRITE"
  | XS_MKDIR -> "MKDIR"
  | XS_RM -> "RM"
  | XS_SET_PERMS -> "SET_PERMS"
  | XS_WATCH_EVENT -> "WATCH_EVENT"
  | XS_ERROR -> "ERROR"
  | XS_IS_DOMAIN_INTRODUCED -> "IS_DOMAIN_INTRODUCED"
  | XS_RESUME -> "RESUME"
  | XS_SET_TARGET -> "SET_TARGET"
  | XS_UNKNOWN -> "UNKNOWN";;

(* C FUNCTIONS *)

external mb : unit -> unit = "mb_c";;

external get_page_size : int -> int32 = "getpagesize_c";;

external mmap : Unix.file_descr -> int -> (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "mmap_c";;

external munmap : (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t -> int32 -> int32 = "munmap_c";;

external xc_domain_getinfo : Unix.file_descr -> int32 -> int -> (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t -> int32 = "xc_domain_getinfo_c";;

external xc_map_foreign_range : int -> int -> int32 -> int -> (char, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "xc_map_foreign_range_c";;

external xc_interface_close : int -> int32 = "xc_interface_close";;

external xc_event_chan_bind_interdomain : Unix.file_descr -> int -> int32 -> int32 = "xc_evtchn_bind_interdomain_c";;

external xc_event_chan_bind_virq : Unix.file_descr -> int -> int32 = "xc_evtchn_bind_virq_c";;

(* Return event channel *)
external xc_event_chan_fd : Unix.file_descr -> int32 = "xc_evtchn_fd_c";;

external xc_event_chan_notify : Unix.file_descr -> int32 -> int32 = "xc_evtchn_notify_c";;

external xc_event_chan_open : unit -> int32 = "xc_evtchn_open_c";;

external xc_event_chan_pending : int -> int32 = "xc_evtchn_pending_c";;

external xc_event_chan_unbind : Unix.file_descr -> int32 -> int32 = "xc_evtchn_unbind_c";;

external xc_event_chan_unmask : int -> int32 -> int32 = "xc_evtchn_unmask_c";;

external xc_interface_open : unit -> int32 = "xc_interface_open_c";;

(* XENBUS FUNCTIONS *)

(* XXX: This is to force libxenctrl to be compiled in *)
let fake () =
  xc_interface_close 0;;

(* xenstored proc files *)
let xenstored_proc_kva = "/proc/xen/xsd_kva";;
let xenstored_proc_port = "/proc/xen/xsd_port";;

(* Notify kernel XenStore is running This works for Linux and FreeBSD, but *)
(* not Solaris                                                             *)
let xenbus_notify_running () =
  ();;

let xenbus_event_chan () =
  (* TODO: error opening? *)
  let fd = Unix.openfile xenstored_proc_port [ Unix.O_RDONLY ] 0x0 in
  
  let str = String.create 20 in
  let rc = Unix.read fd str 0 (String.length str) in
  
  if (rc = (- 1)) then (
    Unix.close fd;
    
    (- 1);
  ) else (
    let str2 = String.sub str 0 rc in
    
    let port = int_of_string str2 in
    
    Unix.close fd;
    
    port;
  );;

let xenbus_map () =
  let fd = Unix.openfile xenstored_proc_kva [ Unix.O_RDWR ] 0o600
  and page_size = Int32.to_int (get_page_size 0) in
  
  let interface = mmap fd page_size in
  
  Unix.close fd;
  
  interface;;

(* PERMISSION FUNCTIONS *)

(* Convert a XenStore permission to a string *)
let xs_perm_to_string perm =
  let perm_str =
    match !(perm.perms) with
    | [ XS_PERM_WRITE ] -> "w"
    | [ XS_PERM_READ ] -> "r"
    | [ XS_PERM_READ; XS_PERM_WRITE ] -> "b"
    | [ XS_PERM_WRITE; XS_PERM_READ ] -> "b"
    | [ XS_PERM_NONE ] -> "n"
    | _ -> raise (Unix.Unix_error (Unix.EINVAL, "xs_perm_to_string", "Invalid permission type"))
  in
  
  perm_str ^ (string_of_int !(perm.perm_dom_id));;

(* Convert a list of XenStore permissions to a null-separated string *)
let rec perms_to_strings perms =
  if ((List.length perms) = 0) then (
    String.create 0;
  ) else (
    null_terminate (xs_perm_to_string (List.hd perms)) ^ perms_to_strings (List.tl perms);
  );;

(* PRINT FUNCTIONS *)

(* Print an error to the error stream and die *)
let barf s =
  Printf.eprintf "FATAL: %s\n" s;
  exit 1;;

(* Print an error to the error stream and die *)
let barf_perror s =
  barf s;;

(* Parse command-line options *)
let parse_error _ =
  barf (Printf.sprintf "%s: No arguments desired\n" Sys.argv.(0));;

(* Helper print store function *)
let rec print_store2 indent store =
  Printf.printf "%s%s: perms: [ %s ]" indent store.name (perms_to_strings !(store.permissions));
  
  ignore (
      match !(store.contents) with
      | Value (v) -> Printf.printf " data: \"%s\"\n" v
      | Children (c) -> (
            ignore (
                match !(store.contents2) with
                | Some (cont) -> (
                      match cont with
                      | Value (v) -> Printf.printf " data: \"%s\"\n" v
                      | Children (c2) -> (); (* XXX: This should never happen... *)
                    )
                | None -> ();
              );
            Printf.printf "\n";
            List.iter (print_store2 (indent ^ "  ")) c;
          );
    );
  ();;

(* Print a store *)
let print_store store =
  print_store2 "" store;;

(* CONNECTION FUNCTIONS *)

(* Create a new input buffer *)
let new_buffer () =
  let (buff : buffered_data) = {
    in_header = ref true;
    used = ref 0;
    header = ref {
          msg_type = ref XS_UNKNOWN;
          req_id = ref 0L;
          tx_id = ref 0L;
          len = ref 0L;
          buff = (String.create header_size);
        };
    buffer = ref (String.create 0);
  } in
  
  buff;;

(* Create a new connection *)
let new_connection write read =
  let (conn : connection) = {
    fd = ref None;
    id = ref 0;
    
    can_write = ref true;
    
    in_buff = ref (new_buffer ());
    out_list = ref [];
    
    transaction = ref None;
    
    transaction_list = ref [];
    next_transaction_id = ref 0L;
    transaction_started = ref 0L;
    
    domain = ref None;
    
    watches = ref [];
    
    write = ref write;
    read = ref read;
  } in
  
  connections := conn :: !connections;
  
  (conn);;

(* TRANSACTION FUNCTIONS *)

(* Add a change node for a transaction *)
let add_change_node trans name recurse =
  match trans with
  | None -> (
        (* Changing global store *)
        generation := Int64.succ !generation;
      )
  | Some (t) -> (
        (* Check if node already is in list *)
        let f n = (
            n.node_name = name;
          )
        in
        
        if (not (List.exists f !(t.changes))) then (
          let new_changed_node = {
            node_name = name;
            recurse = recurse;
          } in
          
          t.changes := List.append !(t.changes) [ new_changed_node ];
        );
      );;

(* The null transaction *)
let null_transaction = {
  trans_id = ref Int64.zero;
  generation = ref Int64.zero;
  store = ref {
        name = "";
        contents = ref (Value (""));
        
        (* XXX: HUGE HACK for broken tools *)
        contents2 = ref None;
        
        permissions = ref [ { perm_dom_id = ref 0; perms = ref [ XS_PERM_NONE ] } ];
      };
  changes = ref [];
  changed_domains = ref [];
};;

(* Look up a transaction for a connection *)
let transaction_lookup conn id =
  if (id <> Int64.zero) then (
    let f trans = (
        !(trans.trans_id) = id;
      )
    in
    
    try (
        Some (List.nth (List.filter f !(conn.transaction_list)) 0);
      )
    with e -> (
          None;
        );
  ) else (
    None;
  );;

(* Get an unused transaction ID *)
let rec get_unused_transaction_id conn =
  let id = !(conn.next_transaction_id) in
  let exists = transaction_lookup conn id in
  
  conn.next_transaction_id := Int64.succ id;
  
  match exists with
  | Some (t) -> get_unused_transaction_id conn
  | None -> (
        if (id = Int64.zero) then (
          get_unused_transaction_id conn;
        ) else (
          id;
        );
      );;

(* Return a deep copy of a list *)
let deep_list_copy lst =
  let rec loop i = (
      if (i = List.length lst) then (
        [];
      ) else (
        List.nth lst i :: loop (succ i);
      );
    )
  in
  
  loop 0;;

(* Return a deep copy of a store tree *)
let rec deep_node_copy node =
  let new_node = {
    name = node.name;
    contents = ref (
          match !(node.contents) with
          | Value (v) -> Value (String.copy v)
          | Children (c) -> (
                let rec loop i = (
                    if (i < (List.length c)) then (
                      (deep_node_copy (List.nth c i)) :: (loop (succ i));
                    ) else (
                      [];
                    );
                  )
                in
                
                Children (loop 0);
              );
        );
    
    (* XXX: HUGE HACK for broken tools *)
    contents2 = ref !(node.contents2);
    
    permissions = ref (deep_list_copy !(node.permissions));
  } in
  
  new_node;;

let transaction_entries_inc transaction dom_id =
  let f domain = (
      domain.changed_dom_id = dom_id;
    )
  in
  
  try (
      let domain = List.find f !(transaction.changed_domains) in
      
      incr (domain.changed_num_entries);
    )
  with e -> (
        match e with
        | Not_found -> (
              let changed_domain = { changed_dom_id = dom_id; changed_num_entries = ref (- 1); } in
              
              transaction.changed_domains := !(transaction.changed_domains) @ [ changed_domain ];
            )
        | _ -> raise e;
      );;

let transaction_entries_dec transaction dom_id =
  let f domain = (
      domain.changed_dom_id = dom_id;
    )
  in
  
  try (
      let domain = List.find f !(transaction.changed_domains) in
      
      decr (domain.changed_num_entries);
    )
  with e -> (
        match e with
        | Not_found -> (
              let changed_domain = { changed_dom_id = dom_id; changed_num_entries = ref (- 1); } in
              
              transaction.changed_domains := !(transaction.changed_domains) @ [ changed_domain ];
            )
        | _ -> raise e;
      );;

(* DOMAIN FUNCTIONS *)

(* Event channel *)
let (xce_handle : Unix.file_descr option ref) = ref None;;

let virq_port = ref (- 1);;
let xc_handle = ref (- 1);;

let virq_dom_exc = 3;;

(* Convert a string of bytes into an int *)
let bytes_to_int bytes =
  let temp = ref Int64.zero in
  let size = 4 in
  
  (* Convert bytes to an int *)
  let rec convert_loop buff i ofs = (
      if (i < size) then (
        temp := Int64.add (Int64.mul !temp 256L) (Int64.of_int (int_of_char (buff.[(size - 1 - i) + ofs])));
        
        convert_loop buff (succ i) ofs;
      );
    ) in
  
  convert_loop bytes 0 0;
  !temp;;

(* Convert an int into a string of bytes *)
let int_to_bytes x =
  let size = 4 in
  let buff = String.create size in
  
  (* Convert bytes to an int *)
  let rec convert_loop i ofs n = (
      if (i < size) then (
        buff.[ofs + i] <- char_of_int (Int64.to_int (Int64.rem n 256L));
        
        convert_loop (succ i) ofs (Int64.div n 256L);
      );
    ) in
  
  (* Convert header entries to ints *)
  convert_loop 0 0 x;
  buff;;

(* Copy a portion of a Bigarray to a string *)
let arr_to_string src src_ofs dest dest_ofs len =
  let rec loop i = (
      if (i < len) then (
        dest.[dest_ofs + i] <- src.{ src_ofs + i };
        
        loop (succ i);
      );
    )
  in
  
  loop 0;;

(* Copy a string into a Bigarray *)
let string_to_arr src src_ofs dest dest_ofs len =
  let rec loop i = (
      if (i < len) then (
        dest.{ dest_ofs + i } <- src.[src_ofs + i];
        
        loop (succ i);
      );
    )
  in
  
  loop 0;;

(* Return an interface's request ring buffer *)
let get_req intf =
  let buff = String.create xenstore_ring_size in
  
  arr_to_string intf 0 buff 0 xenstore_ring_size;
  buff;;

(* Return an interface's response ring buffer *)
let get_rsp intf =
  let buff = String.create xenstore_ring_size in
  
  arr_to_string intf xenstore_ring_size buff 0 xenstore_ring_size;
  buff;;

(* Return an int from the given offset into a Bigarray *)
let get_int a ofs =
  let size = 4 in
  let bytes = String.create size in
  
  arr_to_string a ofs bytes 0 size;
  bytes_to_int bytes;;

(* Return an interface's request consumer index *)
let get_req_cons intf =
  let ofs = xenstore_ring_size * 2 in
  
  get_int intf ofs;;

(* Return an interface's request producer index *)
let get_req_prod intf =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + size in
  
  get_int intf ofs;;

(* Return an interface's response consumer index *)
let get_rsp_cons intf =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + (size * 2) in
  
  get_int intf ofs;;

(* Return an interface's response producer index *)
let get_rsp_prod intf =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + (size * 3) in
  
  get_int intf ofs;;

(* Set an int at the given offset into a Bigarray *)
let set_int a ofs n =
  let size = 4 in
  let bytes = int_to_bytes n in
  
  string_to_arr bytes 0 a ofs size;;

(* Set an interface's request consumer index *)
let set_req_cons intf idx =
  let ofs = xenstore_ring_size * 2 in
  
  set_int intf ofs idx;;

(* Set an interface's request producer index *)
let set_req_prod intf idx =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + size in
  
  set_int intf ofs idx;;

(* Set an interface's response consumer index *)
let set_rsp_cons intf idx =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + (size * 2) in
  
  set_int intf ofs idx;;

(* Set an interface's response producer index *)
let set_rsp_prod intf idx =
  let size = 4 in
  let ofs = (xenstore_ring_size * 2) + (size * 3) in
  
  set_int intf ofs idx;;

(* Check if the indexes are valid *)
let check_indexes cons prod =
  Int64.compare (Int64.sub prod cons) (Int64.of_int xenstore_ring_size) <= 0;;

(* Mask a XenStore ring index to be within range *)
let mask_xenstore_idx idx =
  Int64.logand idx (Int64.of_int (xenstore_ring_size - 1));;

(* Return the start index of an output chunk *)
let get_output_chunk cons prod len =
  let ring_size = Int64.of_int xenstore_ring_size
  and start = mask_xenstore_idx prod in
  
  let diff = Int64.sub ring_size (Int64.sub prod cons)
  and left_in_ring = Int64.sub ring_size start in
  
  len := Int64.to_int (
      if ((Int64.compare diff left_in_ring) < 0) then (
        diff;
      ) else (
        left_in_ring;
      ));
  
  Int64.to_int start;;

(* Return the start index of an input chunk *)
let get_input_chunk cons prod len =
  let ring_size = Int64.of_int xenstore_ring_size
  and start = (mask_xenstore_idx cons) in
  
  let diff = Int64.sub prod cons
  and left_in_ring = Int64.sub ring_size start in
  
  len := Int64.to_int (
      if ((Int64.compare diff left_in_ring) < 0) then (
        diff;
      ) else (
        left_in_ring;
      ));
  
  Int64.to_int start;;

(* Copy data into a domain interface's response ring *)
let blit_rsp intf i_ofs data d_ofs len =
  string_to_arr data d_ofs intf (xenstore_ring_size + i_ofs) len;;

(* Copy data into a domain interface's request ring *)
let blit_req intf i_ofs data d_ofs len =
  arr_to_string intf i_ofs data d_ofs len;;

(* Write to event channel *)
let writechn conn data ofs len =
  match !(conn.domain) with
  | Some (d) -> (
        match !(d.interface) with
        | Some (intf) -> (
              match !xce_handle with
              | Some (xce_h) -> (
                    let cons = get_rsp_cons intf
                    and prod = get_rsp_prod intf in
                    
                    mb ();
                    
                    if (not (check_indexes cons prod)) then (
                      raise (Unix.Unix_error (Unix.EIO, "writechn", "could not check indexes"));
                    ) else (
                      let avail = ref 0 in
                      let dest_start = get_output_chunk cons prod avail in
                      
                      let len2 =
                        if (!avail < len) then (
                          !avail;
                        ) else (
                          len;
                        ) in
                      
                      blit_rsp intf dest_start data ofs len2;
                      
                      mb ();
                      
                      set_rsp_prod intf (Int64.add (get_rsp_prod intf) (Int64.of_int len2));
                      
                      ignore (xc_event_chan_notify xce_h (Int32.of_int (int_of_file_descr !(d.port))));
                      
                      len2;
                    );
                  )
              | None -> (- 1);
            )
        | None -> (- 1);
      )
  | None -> (- 1);;

(* Read from event channel *)
let readchn conn data ofs len =
  match !(conn.domain) with
  | Some (d) -> (
        match !(d.interface) with
        | Some (intf) -> (
              match !xce_handle with
              | Some (xce_h) -> (
                    let cons = get_req_cons intf
                    and prod = get_req_prod intf in
                    
                    mb ();
                    
                    if (not (check_indexes cons prod)) then (
                      raise (Unix.Unix_error (Unix.EIO, "readchn", "could not check indexes"));
                    ) else (
                      let avail = ref 0 in
                      let src_start = get_input_chunk cons prod avail in
                      
                      let len2 =
                        if (!avail < len) then (
                          !avail;
                        ) else (
                          len;
                        ) in
                      
                      blit_req intf src_start data ofs len2;
                      
                      mb ();
                      
                      set_req_cons intf (Int64.add (get_req_cons intf) (Int64.of_int len2));
                      
                      ignore (xc_event_chan_notify xce_h (Int32.of_int (int_of_file_descr !(d.port))));
                      
                      len2;
                    );
                  )
              | None -> (- 1);
            )
        | None -> (- 1);
      )
  | None -> (- 1);;

(* New domain *)
let new_domain dom_id port =
  (* Tell kernel we're interested in this event *)
  let rc =
    match !xce_handle with
    | Some (fd) -> (
          Int32.to_int (xc_event_chan_bind_interdomain fd dom_id (Int32.of_int (int_of_file_descr port)));
        )
    | None -> (
          (- 1);
        );
  in
  
  if (rc <> (- 1)) then (
    let domain = {
      dom_id = ref dom_id;
      
      port = ref (file_descr_of_int rc);
      remote_port = ref port;
      
      mfn = ref 0;
      path = ref ("/local/domain/" ^ (string_of_int dom_id));
      
      interface = ref None;
      
      conn = ref (new_connection writechn readchn);
      
      shutdown = ref false;
      
      num_entries = ref 0;
      num_watches = ref 0;
    } in
    
    !(domain.conn).domain := Some (domain);
    !(domain.conn).id := dom_id;
    (* domain.port := file_descr_of_int rc; *)
    
    (* Add to list of domains *)
    domains := !domains @ [ domain ];
    
    Some (domain);
  ) else (
    None;
  );;

(* Domain-0 initialisation *)
let dom0_init () =
  let port = xenbus_event_chan () in
  
  if (port <> (- 1)) then (
    let dom0 = new_domain 0 (file_descr_of_int port) in
    
    match dom0 with
    | Some (d) -> (
          d.interface := Some (xenbus_map ());
          
          match !(d.interface) with
          | Some (i) -> (
                ignore (
                    match !xce_handle with
                    | Some (fd) -> ignore (xc_event_chan_notify fd (Int32.of_int (int_of_file_descr !(d.port))))
                    | None -> ();
                  );
                
                0;
              )
          | None -> (
                (- 1);
              );
        )
    | None -> (
          (- 1);
        );
  ) else (
    (- 1);
  );;

(* Initialise domain *)
let domain_init () =
  let rc = ref 0 in
  
  xc_handle := Int32.to_int (xc_interface_open ());
  
  if (!xc_handle < 0) then (
    barf_perror "Failed to open connection to hypervisor\n";
  );
  
  let xce_handle_int = Int32.to_int (xc_event_chan_open ()) in
  
  if (xce_handle_int < 0) then (
    barf_perror "Failed to open evtchn device\n";
  ) else (
    xce_handle := Some (file_descr_of_int xce_handle_int);
  );
  
  if ((dom0_init ()) <> 0) then (
    barf_perror "Failed to initialise dom0 state\n";
  );
  
  rc := Int32.to_int (xc_event_chan_bind_virq (file_descr_of_int xce_handle_int) virq_dom_exc);
  
  if (!rc = (- 1)) then (
    barf_perror "Failed to bind to domain exception virq port\n";
  );
  
  virq_port := !rc;
  
  !xce_handle;;

(* Check if the domain can read *)
let domain_can_read conn =
  match !(conn.domain) with
  | Some (d) -> (
        match !(d.interface) with
        | Some (intf) -> (Int64.compare (get_req_cons intf) (get_req_prod intf)) <> 0
        | None -> false;
      )
  | None -> false;;

(* Check if the domain can write *)
let domain_can_write conn =
  match !(conn.domain) with
  | Some (d) -> (
        match !(d.interface) with
        | Some (intf) -> (Int64.compare (Int64.sub (get_rsp_prod intf) (get_rsp_cons intf)) (Int64.of_int xenstore_ring_size)) <> 0
        | None -> false;
      )
  | None -> false;;

(* Find a domain based on its ID *)
let find_domain_by_domid dom_id =
  let f d = (
      !(d.dom_id) = dom_id;
    )
  in
  
  try (
      Some (List.find f !domains);
    )
  with e -> (
        match e with
        | Not_found -> None
        | _ -> raise e;
      );;

(* Fix a domain entry *)
let domain_entry_fix dom_id num =
  match find_domain_by_domid dom_id with
  | Some (dom) -> (
        dom.num_entries := !(dom.num_entries) + num;
        
        if (!(dom.num_entries) < 0) then (
          dom.num_entries := 0;
        );
      )
  | None -> ();;

(* Initialise dominfo *)
let dominfo_init dominfo =
  let size = 4 in
  let ofs = 8 in
  
  let rec loop i ofs num = (
      if (i < num) then (
        set_int dominfo (ofs + (i * size)) Int64.one;
        loop (succ i) ofs num;
      );
    )
  in
  
  loop 0 ofs 8;
  ();;

(* Return the dominfo dom id *)
let get_dominfo_domid dominfo =
  get_int dominfo 0;;

(* Return the dominfo dying status *)
let get_dominfo_dying dominfo =
  let ofs = 8 in
  (get_int dominfo ofs) <> Int64.zero;;

(* Return the dominfo crashed status *)
let get_dominfo_crashed dominfo =
  let ofs = 12 in
  (get_int dominfo ofs) <> Int64.zero;;

(* Return the dominfo shutdown status *)
let get_dominfo_shutdown dominfo =
  let ofs = 16 in
  (get_int dominfo ofs) <> Int64.zero;;

(* Return if a domain is unprivileged *)
let domain_is_unprivileged conn =
  match !(conn.domain) with
  | Some (d) -> !(d.dom_id) <> 0
  | None -> false;;

(* Return the number of entries left for a domain *)
let get_domain_entries conn =
  if (domain_is_unprivileged conn) then (
    match !(conn.domain) with
    | Some (d) -> !(d.num_entries)
    | None -> 0; (* XXX: This should never happen... *)
  ) else (
    0;
  );;

let domain_entries_inc conn node =
  let node_owner_id = !((List.hd !(node.permissions)).perm_dom_id) in
  
  if (node_owner_id <> !(conn.id)) then (
    match !(conn.transaction) with
    | Some (transaction) -> transaction_entries_inc transaction node_owner_id
    | None -> (
          match (find_domain_by_domid node_owner_id) with
          | Some (domain) -> incr (domain.num_entries)
          | None -> ();
        );
  ) else (
    match !(conn.domain) with
    | Some (domain) -> (
          match !(conn.transaction) with
          | Some (transaction) -> transaction_entries_inc transaction !(domain.dom_id)
          | None -> incr (domain.num_entries);
        )
    | None -> ();
  );;

let domain_entries_dec conn node =
  let node_owner_id = !((List.hd !(node.permissions)).perm_dom_id) in
  
  if (node_owner_id <> !(conn.id)) then (
    match !(conn.transaction) with
    | Some (transaction) -> transaction_entries_dec transaction node_owner_id
    | None -> (
          match (find_domain_by_domid node_owner_id) with
          | Some (domain) -> (
                if (!(domain.num_entries) <> 0) then (
                  decr (domain.num_entries);
                );
              )
          | None -> ();
        );
  ) else (
    match !(conn.domain) with
    | Some (domain) -> (
          if (!(domain.num_entries) <> 0) then (
            match !(conn.transaction) with
            | Some (transaction) -> transaction_entries_inc transaction !(domain.dom_id)
            | None -> decr (domain.num_entries);
          );
        )
    | None -> ();
  );;

(* XENSTORE FUNCTIONS *)

(* Return the XenStore daemon run directory *)
let xs_daemon_rundir () =
  let s = ref "" in
  
  ignore (
      try
        s := Sys.getenv "XENSTORED_RUNDIR"
      
      with e -> (
            match e with
            | Not_found -> s := "/var/run/xenstored"
            | _ -> raise e;
          );
    );
  
  !s;;

(* Trace I/O *)
let trace_io conn prefix data =
  match !trace_fd with
  | Some (tfd) -> (
        let tm = Unix.localtime (Unix.gettimeofday ()) in
        
        (* Create trace string *)
        let conn_fd =
          match !(conn.fd) with
          | Some (fd) -> int_of_file_descr fd
          | None -> (- 1);
        in
        
        let trace_str = ref (Printf.sprintf "%s %d %04d%02d%02d %02d:%02d:%02d %s (" prefix conn_fd (tm.Unix.tm_year + 1900) (tm.Unix.tm_mon + 1) tm.Unix.tm_mday tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec (sockmsg_string !(!(data.header).msg_type))) in
        
        (* FIXME: Clean this up *)
        let rec loop i = (
            let c = ref ' ' in
            
            if (i < (String.length !(data.buffer))) then (
              if (not ((int_of_char !(data.buffer).[i]) = 0)) then (
                c := !(data.buffer).[i];
              );
              
              trace_str := !trace_str ^ (String.make 1 !c);
              loop (succ i);
            );
          )
        in
        
        loop 0;
        
        trace_str := !trace_str ^ ")\n";
        
        (* TODO: Write to tfd *)
        Printf.printf "%s" !trace_str;
        flush stdout;
      )
  | None -> ();;

(* Return the XenStore daemon root directory *)
let xs_daemon_rootdir () =
  let s = ref "" in
  
  ignore (
      try
        s := Sys.getenv "XENSTORED_ROOTDIR"
      
      with e -> (
            match e with
            | Not_found -> s := "/var/lib/xenstored"
            | _ -> raise e;
          );
    );
  
  !s;;

(* Return the XenStore daemon path *)
let xs_daemon_path () =
  let s = ref "" in
  
  ignore (
      try
        s := Sys.getenv "XENSTORED_PATH"
      
      with e -> (
            match e with
            (* TODO: Check >= PATH_MAX *)
            | Not_found -> s := (xs_daemon_rundir () ^ "/socket")
            | _ -> raise e;
          );
    );
  
  !s;;

(* Return the name of the XenStore socket *)
let xs_daemon_socket () =
  xs_daemon_path ();;

(* Return the name of the XenStore read-only socket *)
let xs_daemon_socket_ro () =
  (xs_daemon_path ()) ^ "_ro";;

(* Create the given xenstored directory, if needed *)
let xs_mkdir name =
  if (not (Sys.file_exists name)) then (
    Unix.mkdir name 0o0755;
  );;

(* Remove the old sockets *)
let xs_unlink addr =
  (* FIXME: Be more sophisticated -- Don't mug running daemon *)
  match addr with
  | Unix.ADDR_UNIX (name) -> (
        if (Sys.file_exists name) then
          Unix.unlink name;
      )
  | _ -> Printf.printf "addr -- unlink oops\n";;

(* Change the permissions for a socket address *)
let xs_chmod addr =
  match addr with
  | Unix.ADDR_UNIX (name) -> Unix.chmod name 0o0600
  | _ -> Printf.printf "addr -- chmod oops\n";;

(* Write PID file *)
let write_pid_file pid_file =
  let fd = Unix.openfile pid_file [ Unix.O_RDWR; Unix.O_CREAT ] 0x0600 in
  
  (* Exit silently if daemon already running *)
  ignore (
      try (
          Unix.lockf fd Unix.F_TLOCK 0
        )
      with e -> (
            exit 0;
          )
    );
  
  let pid = string_of_int (Unix.getpid ()) in
  let len = String.length pid in
  
  ignore (
      try
        if (not ((Unix.write fd pid 0 len) = len)) then
          barf_perror (Printf.sprintf "Writing pid file %s" pid_file)
      
      with e -> barf_perror (Printf.sprintf "Writing pid file %s" pid_file)
    );
  
  Unix.close fd;;

(* Reopen log *)
let reopen_log () =
  if (not (!trace_file = "")) then (
    Printf.printf "reopen_log: %s\n" !trace_file;
    
    ignore (
        match !trace_fd with
        | Some (fd) -> (Unix.close fd; Printf.printf "reopen_log: old log closed\n")
        | None -> ()
      );
    
    trace_fd := Some (Unix.openfile !trace_file [ Unix.O_WRONLY; Unix.O_CREAT; Unix.O_APPEND ] 0o0600);
  );;

(* TODO: Trigger reopen log *)
let trigger_reopen_log signal =
  ();;

(* Try to fork daemon *)
let fork_daemon () =
  let pid = Unix.fork () in
  
  if (pid < 0) then (
    barf "Failed to fork daemon";
  );
  
  if (pid <> 0) then (
    exit 0;
  );
  
  pid;;

(* Daemonise *)
let daemonise () =
  (* Separate from parent via fork, so init inherits us *)
  let pid = ref (fork_daemon ()) in
  
  (* Session leader so ^C doesn't whack us *)
  ignore (Unix.setsid ());
  
  (* Let session leader exit so child cannot regain CTTY *)
  pid := fork_daemon ();
  
  (* Move off any mount points we might be in *)
  ignore (
      try
        Unix.chdir "/"
      
      with e -> barf "Failed to chdir"
    );
  
  (* Discard parent's old-fashioned umask prejudices *)
  ignore (Unix.umask 0);;

(* FD FUNCTIONS *)

(* Check if a file descriptor exists in a set of descriptors *)
let fd_is_set fd set =
  List.mem fd set;;

(* Add a file descriptor in the set *)
let set_fd fd set =
  (* Check if fd already in set *)
  if (not (fd_is_set fd set)) then (
    fd :: set;
  ) else (
    set;
  );;

(* Initialise file descriptor sets *)
let initialise_set in_set out_set socket socket_ro timeout =
  timeout := (- 1.0);
  
  (* Clear sets *)
  in_set := [];
  out_set := [];
  
  (* Add sockets to the in set *)
  in_set := set_fd socket !in_set;
  in_set := set_fd socket_ro !in_set;
  
  (* TODO: set_fd reopen_log_pipe[0] !in_set; *)
  
  ignore (
      match !xce_handle with
      | Some (xce_h) -> in_set := set_fd xce_h !in_set
      | None -> ();
    );
  
  (* Iterate over all connections *)
  let f conn = (
      match !(conn.domain) with
      | None -> (
            match !(conn.fd) with
            | Some (fd) -> (
                  in_set := set_fd fd !in_set;
                  
                  if (List.length !(conn.out_list) <> 0) then (
                    out_set := set_fd fd !out_set;
                  );
                )
            | None -> ();
          )
      | _ -> (
            if ((domain_can_read conn) || ((domain_can_write conn) && (List.length !(conn.out_list) <> 0))) then (
              timeout := 0.0;
            );
          );
    )
  in
  
  List.iter f !connections;;

(* Read from a connection *)
let readfd conn data ofs len =
  let rc = ref 0 in
  
  ignore (
      match !(conn.fd) with
      | Some (fd) -> (
            let rec loop () = (
                ignore (
                    try (
                        rc := Unix.read fd data ofs len;
                      ) with e -> (
                          match e with
                          | Unix.Unix_error (err, _, _) -> (
                                match err with
                                | Unix.EAGAIN -> rc := 0
                                | Unix.EINTR -> loop ()
                                | _ -> ();
                              );
                          | _ -> (); (* XXX: Barf here instead? *)
                        );
                  );
              ) in
            
            loop ();
            
            (* Reading zero length means we're done with this connection. *)
            if ((!rc = 0) && (len <> 0)) then (
              rc := (- 1);
            );
          )
      | None -> rc := (- 1);
    );
  
  !rc;;

(* Write file descriptor function *)
let writefd conn data ofs len =
  let rc = ref (- 1) in
  
  ignore (
      match !(conn.fd) with
      | Some (fd) -> (
            
            let rec loop () = (
                ignore (
                    try (
                        rc := Unix.single_write fd data ofs len
                      )
                    with e -> (
                          match e with
                          | Unix.Unix_error (Unix.EAGAIN, _, _) -> rc := 0
                          | Unix.Unix_error (Unix.EINTR, _, _) -> (
                                if (!rc < 0) then (
                                  loop ();
                                );
                              )
                          | _ -> (); (* XXX: Barf here instead? *)
                        );
                  );
              ) in
            
            loop ();
          )
      | None -> ();
    );
  
  !rc;;

(* Null connection (tools) *)
let null_conn = {
  fd = ref None;
  id = ref (- 1);
  
  can_write = ref true;
  
  in_buff = ref (new_buffer ());
  out_list = ref [];
  
  transaction = ref None;
  transaction_list = ref [];
  next_transaction_id = ref Int64.zero;
  transaction_started = ref Int64.zero;
  
  domain = ref None;
  
  watches = ref [];
  
  write = ref writefd;
  read = ref readfd;
};;

(* Restore existing XenStore connections *)
let restore_existing_connections () =
  ();;

(* Write messages *)
let write_messages_body conn out n =
  let ret = ref 0 in
  
  ret := (!(conn.write) conn !(out.buffer) !(out.used) ((Int64.to_int !(!(out.header).len)) - !(out.used)));
  
  if (!ret < 0) then (
    false;
  ) else (
    out.used := !(out.used) + !ret;
    
    if (!(out.used) = (Int64.to_int !(!(out.header).len))) then (
      trace_io conn "OUT" out;
      conn.out_list := remove_from_list !(conn.out_list) n;
    );
    
    true;
  );;

let write_messages conn =
  let n = 0 in
  let out = List.nth !(conn.out_list) n in
  
  if (!(out.in_header)) then (
    if (!verbose) then (
      Printf.printf "Writing msg %s (%s) out to %d\n" (sockmsg_string !(!(out.header).msg_type)) (String.sub !(out.buffer) 0 (Int64.to_int !(!(out.header).len))) !(conn.id);
      flush stdout;
    );
    
    let ret = (!(conn.write) conn (!(out.header).buff) !(out.used) (header_size - !(out.used))) in
    
    if (ret < 0) then (
      false;
    ) else (
      out.used := !(out.used) + ret;
      
      if (!(out.used) < header_size) then (
        true;
      ) else (
        out.in_header := false;
        out.used := 0;
        
        (* Second write might block if non-zero *)
        if ((!(!(out.header).len) > 0L) && (!(conn.domain) = None)) then (
          true;
        ) else (
          write_messages_body conn out n;
        );
      );
    );
  ) else (
    write_messages_body conn out n;
  );;

(* Destroy connection *)
let destroy_conn conn =
  ignore (
      match !(conn.fd) with
      | Some (fd) -> (
            (* Flush outgoing if possible, but don't block. *)
            if (!(conn.domain) = None) then (
              let set = set_fd fd [] in
              
              let rec loop () = (
                  if ((List.length !(conn.out_list)) > 0) then (
                    let select_sets = Unix.select [] set [] 0.0 in
                    let out_set = tuple_second select_sets in
                    
                    if ((List.length out_set) = 1) then (
                      if (write_messages conn) then (
                        loop ();
                      );
                    );
                  );
                )
              in
              
              loop ();
              
              Unix.close fd;
            );
          )
      | None -> ();
    );
  
  (* TODO: trace_destroy conn "connection"; *)
  
  (* TODO: Write messages? while (!list_empty(& conn -> out_list) &&       *)
  (* select(conn -> fd + 1, NULL, & set, NULL, & none) == 1) if            *)
  (* (!write_messages(conn)) break;                                        *)
  ();;

(* Remove a connection *)
let remove_connection conn =
  (* XXX: Debug *)
  Printf.printf "Remove connection: id: %d\n" !(conn.id);
  flush stdout;
  
  let (new_connections : connection list ref) = ref [] in
  
  let rec loop i = (
      if (i < List.length !connections) then (
        if (not ((List.nth !connections i) == conn)) then (
          new_connections := List.append !new_connections ((List.nth !connections i) :: []);
        );
        
        loop (succ i);
      );
    ) in
  
  loop 0;
  connections := !new_connections;;

let accept_connection socket can_write =
  let (fd, _) = Unix.accept socket in
  let conn = new_connection writefd readfd in
  
  Printf.printf "Connection on %d\n" (int_of_file_descr fd);
  flush stdout;
  
  conn.fd := Some (fd);
  conn.can_write := can_write;;

(* Populate header variables from buffer *)
let populate_header_variables hdr buff =
  let size = 4 in
  let temp = ref Int64.zero in
  
  (* Convert bytes to an int *)
  let rec convert_loop i ofs = (
      if (i < size) then (
        temp := Int64.add (Int64.mul !temp 256L) (Int64.of_int (int_of_char (buff.[(size - i - 1) + ofs])));
        
        convert_loop (succ i) ofs;
      );
    ) in
  
  (* Convert header entries to ints *)
  convert_loop 0 0;
  hdr.msg_type := int_to_sockmsg_type !temp;
  
  temp := 0L;
  convert_loop 0 4;
  hdr.req_id := !temp;
  
  temp := 0L;
  convert_loop 0 8;
  hdr.tx_id := !temp;
  
  temp := 0L;
  convert_loop 0 12;
  hdr.len := !temp;;

(* Populate reply header *)
let populate_header_buffer header =
  let size = 4 in
  
  let temp = ref Int64.zero in
  let buff = header.buff in
  
  (* Convert bytes to an int *)
  let rec convert_loop i ofs = (
      if (i < size) then (
        String.set buff (i + ofs) (char_of_int (Int64.to_int (Int64.rem !temp 256L)));
        temp := Int64.div !temp 256L;
        convert_loop (succ i) ofs;
      );
    ) in
  
  (* Convert header entries to ints *)
  temp := sockmsg_type_to_int !(header.msg_type);
  convert_loop 0 0;
  
  temp := !(header.req_id);
  convert_loop 0 4;
  
  temp := !(header.tx_id);
  convert_loop 0 8;
  
  temp := !(header.len);
  convert_loop 0 12;;

(* SEND FUNCTIONS *)

(* Send a reply *)
let send_reply conn msg_type data len =
  let bdata = ref (new_buffer ()) in
  let header = !((!bdata).header) in
  
  (* XS_WATCH_EVENT *)
  if (msg_type = XS_WATCH_EVENT) then (
    header.req_id := 0L;
    header.tx_id := 0L;
  ) else (
    header.req_id := !(!(!(conn.in_buff).header).req_id);
    header.tx_id := !(!(!(conn.in_buff).header).tx_id);
  );
  
  header.msg_type := msg_type;
  header.len := (Int64.of_int len);
  populate_header_buffer header;
  
  (!bdata).buffer := data;
  
  conn.out_list := List.append !(conn.out_list) [ !bdata ];;

(* Send an ACK message *)
let send_ack conn msg_type =
  let data = null_terminate "OK" in
  
  send_reply conn msg_type data (String.length data);;

(* Send an error message *)
let send_error conn error =
  let error_string = null_terminate (get_xsd_error_string error) in
  
  send_reply conn XS_ERROR error_string (String.length error_string);;

(* Return if a string contains another string *)
let rec strstr s1 s2 =
  try (
      let i = String.index s1 s2.[0] in
      
      if ((String.length (String.sub s1 i ((String.length s1) - i))) < (String.length s2)) then (
        false;
      ) else (
        let temp = String.sub s1 i (String.length s2) in
        
        if (temp = s2) then (
          true;
        ) else (
          strstr (String.sub s1 (i + 1) ((String.length s1) - i - 1)) s2;
        );
      );
    )
  with e -> (
        match e with
        | Not_found -> false
        | _ -> raise e;
      );;

(* Return the implicit path of a connection (only domains have this) *)
let get_implicit_path conn =
  match !(conn.domain) with
  | Some (d) -> Some !(d.path)
  | None -> None;;

(* Return canonicalised name *)
let canonicalise conn name =
  if ((String.get name 0) = '/') then (
    name;
  ) else (
    let prefix = get_implicit_path conn in
    
    match prefix with
    | Some (s) -> s ^ "/" ^ name
    | None -> name;
  );;

(* Check that a node name contains only valid characters *)
let valid_chars name =
  (* Nodes can have lots of crap *)
  let valid_characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-/_@" in
  
  let rec loop i = (
      if (i < (String.length name)) then (
        if (not (String.contains valid_characters name.[i])) then (
          false;
        ) else (
          loop (succ i);
        );
      ) else (
        true
      );
    )
  
  in
  
  loop 0;;

(* Return if a node name is valid *)
let is_valid_nodename name =
  (* Must start in / *)
  if (name.[0] = '/') then (
    (* Cannot end in / (unless it's just "/") *)
    if ((name = "/") || (name.[(String.length name) - 1] <> '/')) then (
      (* No double // *)
      if (not (strstr name "//")) then (
        (* Make sure it is not too long *)
        if ((String.length name) <= xenstore_abs_path_max) then (
          valid_chars name;
        ) else (
          false;
        );
      ) else (
        false;
      );
    ) else (
      false;
    );
  ) else (
    false;
  );;

(* Return the parent node's name *)
let get_parent_name name =
  try (
      let slash = String.rindex (String.sub name 1 ((String.length name) - 1)) '/' in
      
      String.sub name 0 (slash + 1);
    )
  with e -> (
        match e with
        | Not_found -> "/"
        | _ -> raise e;
      );;

(* Return the basename of a node *)
let get_base_name name =
  if (name = "/") then (
    name;
  ) else (
    let slash = String.rindex name '/' in
    
    String.sub name (slash + 1) ((String.length name) - (slash + 1));
  );;

(* Add a child to a node *)
let add_child_node node child_node =
  (* If it is a leaf node with an empty value, then change the contents to *)
  (* Children                                                              *)
  ignore (
      match !(node.contents) with
      | Value (v) -> (
            if ((String.length v) = 0) then (
              node.contents := Children [];
            )
          )
      | Children (_) -> ();
    );
  
  (* Try to add the child, this only works with non-leaf nodes *)
  ignore (
      match !(node.contents) with
      | Children (c) -> node.contents := Children (List.append c [ child_node ])
      | Value (v) -> (
            Printf.printf "Node: %s -- cannot have both a value and children\n" node.name;
            
            (* XXX: HUGE HACK for broken tools *)
            node.contents2 := Some (Value (v));
            node.contents := Children ([ child_node ]);
          );
    );;

(* Return a child node, if it exists *)
let get_child_node node child_name =
  match !(node.contents) with
  | Children (c) -> (
        let rec loop i = (
            if (i < List.length c) then (
              let child_node = List.nth c i in
              
              if (child_node.name = child_name) then (
                Some child_node;
              ) else (
                loop (succ i);
              );
            ) else (
              None;
            );
          )
        in
        
        loop 0;
      )
  | Value (_) -> None;;

(* Return the store context *)
let get_store_context conn =
  if (!(conn.id) = (- 1)) then (
    !store;
  ) else (
    match !(conn.transaction) with
    | Some (trans) -> (
          if (trans = null_transaction) then (
            !store;
          ) else (
            !(trans.store);
          );
        )
    | None -> !store;
  );;

(* Read a node *)
let rec read_node conn name =
  if (name = "/") then (
    get_store_context conn;
  ) else (
    let parent_node = read_node conn (get_parent_name name) in
    let base_name = get_base_name name in
    
    match (get_child_node parent_node base_name) with
    | Some (n) -> n
    | None -> raise (Unix.Unix_error (Unix.ENOENT, "read_node", name));
  );;

(* Construct a node *)
let rec construct_node conn name =
  let parent_name = get_parent_name name in
  let parent_node = ref None in
  
  ignore (
      try (
          parent_node := Some (read_node conn parent_name);
        )
      with e -> (
            (* If parent doesn't exist, create it *)
            parent_node := construct_node conn parent_name;
          );
    );
  
  match !parent_node with
  | Some (n) -> (
        (* Check quota *)
        if ((get_domain_entries conn) < !quota_num_entry_per_domain) then (
          (* Create new node *)
          let node = {
            name = (get_base_name name);
            contents = ref (Value (String.create 0));
            
            (* XXX: HUGE HACK for broken tools *)
            contents2 = ref None;
            
            (* Inherit permissions *)
            permissions = ref (deep_list_copy !(n.permissions));
          } in
          
          (* Domains own what they create *)
          if ((!(conn.id) <> (- 1)) && (!(conn.id) <> 0)) then (
            (List.hd !(node.permissions)).perm_dom_id := !(conn.id);
          );
          
          (* Add node to parent *)
          add_child_node n node;
          
          Some node;
        ) else (
          None;
        );
      )
  | None -> None;;

(* Create a node *)
let create_node conn name data =
  let node = construct_node conn name in
  
  match node with
  | Some (n) -> (
        n.contents := Value data;
        Some n;
      )
  | None -> None;;

(* Delete a child node *)
let delete_child conn parent child_name =
  match !(parent.contents) with
  | Value (v) -> false
  | Children (c) -> (
        let match_child node = (
            node.name = child_name;
          )
        in
        
        if (List.exists match_child c) then (
          let not_match_child node = (
              node.name <> child_name;
            )
          in
          
          parent.contents := Children (List.filter not_match_child c);
          true;
        ) else (
          (* corrupt(conn, "Can't find child '%s' in %s", childname,       *)
          (* node->name);                                                  *)
          false;
        );
      );;

(* STORE FUNCTIONS *)

(* Setup XenStore *)
let setup_store store =
  ignore (create_node null_conn "/tool/xenstored" "");;

(* Check if a node is a child of another node *)
let is_child child parent =
  if (parent = "/") then (
    true;
  ) else (
    let len = min (String.length parent) (String.length child) in
    
    if ((String.sub child 0 len) <> (String.sub parent 0 len)) then (
      false;
    ) else (
      let len2 = String.length parent in
      
      if (len2 = (String.length child)) then (
        (* XXX: This returns child = parent *)
        true;
      ) else if (len2 < (String.length child)) then (
        String.get child len2 = '/';
      ) else (
        false;
      )
    );
  );;

(* WATCH FUNCTIONS *)

(* Return the number of watches attached to this domain *)
let get_domain_watches conn =
  if (domain_is_unprivileged conn) then (
    match !(conn.domain) with
    | Some (d) -> !(d.num_watches)
    | None -> 0; (* XXX: This should never happen... *)
  ) else (
    0;
  );;

(* Increment the number of watches associated with this domain *)
let domain_watches_inc conn =
  match !(conn.domain) with
  | Some (d) -> incr (d.num_watches)
  | None -> ();;

(* Decrement the number of watches associated with this domain *)
let domain_watches_dec conn =
  match !(conn.domain) with
  | Some (d) -> decr (d.num_watches)
  | None -> ();;

(* Check event node *)
let check_event_node name =
  (* errno = EINVAL? *)
  name.[0] = '@';;

(* Send a watch event *)
let send_event conn watch name =
  let node_name =
    match watch.relative_path with
    | Some (path) -> (
          let n = String.sub name (String.length path) ((String.length name) - (String.length path)) in
          
          if (n.[0] = '/') then (
            String.sub n 1 ((String.length n) - 1);
          ) else (
            n;
          );
        )
    | None -> name;
  in
  
  (* Data to send (node_name\0token\0) *)
  let data = (null_terminate node_name) ^ (null_terminate watch.token) in
  
  send_reply conn XS_WATCH_EVENT data (String.length data);;

(* Get the node permission for a connection *)
let perm_for_conn conn perms =
  let mask = (
      if (not !(conn.can_write)) then (
        [ XS_PERM_READ; XS_PERM_OWNER ];
      ) else (
        [ XS_PERM_READ; XS_PERM_WRITE; XS_PERM_OWNER ];
      );
    ) in
  
  let and_mask perm = (
      List.mem perm mask;
    )
  in
  
  if ((!(conn.id) = 0) || (!((List.hd perms).perm_dom_id) = !(conn.id))) then (
    List.filter and_mask [ XS_PERM_READ; XS_PERM_WRITE; XS_PERM_OWNER];
  ) else (
    if ((List.length perms) > 1) then (
      let f perm = (
          !(perm.perm_dom_id) = !(conn.id)
        )
      in
      
      try (
          List.filter and_mask !((List.find f (List.tl perms)).perms);
        )
      with e -> (
            match e with
            | Not_found -> List.filter and_mask !((List.hd perms).perms)
            | _ -> raise e;
          );
    ) else (
      List.filter and_mask !((List.hd perms).perms);
    );
  );;

(* Ask parents for their permissions *)
let ask_parents conn name =
  let rec loop name = (
      try (
          let node = read_node conn name in
          
          perm_for_conn conn !(node.permissions);
        )
      with e -> (
            if (name <> "/") then (
              loop (get_parent_name name);
            ) else (
              (* TODO: corrupt conn "No permissions file at root" *)
              
              (* loop name; *)
              
              (* XXX: Hack *)
              [ XS_PERM_NONE ];
            );
          );
    )
  in
  
  loop (get_parent_name name);;

(* Return the errno from the parents *)
let errno_from_parents conn name errno perms =
  if (errno = Unix.ENOMEM) then (
    errno;
  ) else (
    let f perm = (
        List.mem perm perms;
      )
    in
    
    if ((List.length (List.filter f (ask_parents conn name))) = (List.length perms)) then (
      errno;
    ) else (
      Unix.EACCES;
    );
  );;

(* Return a node *)
let get_node conn name perms =
  if (is_valid_nodename name) then (
    try (
        let node = read_node conn name in
        
        let and_perms perm = (
            List.mem perm perms;
          )
        in
        
        if ((List.filter and_perms (perm_for_conn conn !(node.permissions))) = perms) then (
          node;
        ) else (
          raise (Unix.Unix_error (Unix.EACCES, "get_node", "Invalid permissions"));
        );
      )
    with e -> (
          match e with
          | Unix.Unix_error (errno, a, b) -> raise (Unix.Unix_error (errno_from_parents conn name errno perms, a, b))
          | _ -> raise e;
        );
  ) else (
    raise (Unix.Unix_error (Unix.EINVAL, "get_node", "Invalid node name"));
  );;

(* Add a watch event *)
let add_event conn watch name =
  if (not (check_event_node name)) then (
    try (
        ignore (
            get_node conn name [ XS_PERM_READ ];
          );
        
        send_event conn watch name;
      )
    with e -> (
          match e with
          | Unix.Unix_error (Unix.ENOENT, _, _) -> send_event conn watch name
          | Unix.Unix_error (Unix.EACCES, _, _) -> send_event conn watch name
          | _ -> ();
        );
    
  ) else (
    send_event conn watch name;
  );;

(* Fire watches *)
let fire_watches conn name recurse =
  (* During transactions, don't fire watches *)
  if (not ((!(conn.id) <> (- 1)) && (!(conn.transaction) <> None))) then (
    (* Create an event for each watch *)
    let f c = (
        let f2 w = (
            if (is_child name w.node) then (
              add_event c w name;
            ) else (
              if (recurse && (is_child w.node name)) then (
                add_event c w w.node;
              );
            );
          )
        in
        
        List.iter f2 !(c.watches);
      )
    in
    
    List.iter f !connections;
  );;

(* UNSORTED FUNCTIONS *)

(* TODO: Domain cleanup *)
let domain_cleanup () =
  let notify = ref false in
  
  let f domain = (
      let size = Int32.to_int (get_page_size 0) in
      let dominfo = Bigarray.Array1.create Bigarray.char Bigarray.c_layout size in
      
      dominfo_init dominfo;
      
      if (((Int32.to_int (xc_domain_getinfo (file_descr_of_int !xc_handle) (Int32.of_int !(domain.dom_id)) 1 dominfo)) = 1) && ((Int64.to_int (get_dominfo_domid dominfo)) = !(domain.dom_id))) then (
        if (((get_dominfo_crashed dominfo) || (get_dominfo_shutdown dominfo)) && (not !(domain.shutdown))) then (
          domain.shutdown := true;
          notify := true;
        );
        
        (* if (not (get_dominfo_dying dominfo)) then ( continue; ); *)
      );
      
      (* destroy_conn domain.conn; destroy_domain domain; notify := false; *)
      (* (* destroy_domain() fires the watch *)                            *)
    )
  in
  
  List.iter f !domains;
  
  if (!notify) then (
    fire_watches null_conn "@releaseDomain" false;
  );;

(* Convert a series of null terminated strings to a list of permissions *)
let rec xs_strings_to_perms str =
  let perm_char = str.[0] in
  let null_index = String.index str (char_of_int 0) in
  let id_str = String.sub str 1 (pred null_index) in
  
  let perm_types =
    match perm_char with
    | 'r' -> [ XS_PERM_READ ]
    | 'w' -> [ XS_PERM_WRITE ]
    | 'b' -> [ XS_PERM_READ; XS_PERM_WRITE ]
    | 'n' -> [ XS_PERM_NONE ]
    | _ -> raise (Unix.Unix_error (Unix.EINVAL, "xs_strings_to_perms", "Invalid permission type"))
  in
  
  let perm = { perm_dom_id = ref (int_of_string id_str); perms = ref perm_types } in
  
  perm :: (
    if ((String.length str) = (succ null_index)) then (
      [];
    ) else (
      (xs_strings_to_perms (String.sub str (succ null_index) ((String.length str) - (succ null_index))));
    );
  );;

(* Write to a node *)
let write_node conn node data perms =
  let over_quota =
    match data with
    | Some (d) -> (String.length d) >= !quota_max_entry_size
    | None -> false
  in
  
  if ((domain_is_unprivileged conn) && over_quota) then (
    raise (Unix.Unix_error (Unix.ENOSPC, "write_node", "Exceded max entry size quota"));
  ) else (
    node.permissions := perms;
    
    match data with
    | Some (d) -> (
          match !(node.contents) with
          | Value (v) -> (
                node.contents := Value (d);
                true;
              )
          | Children (c) -> (
                if ((List.length c) > 0) then (
                  Printf.printf "do_write: node: (%s): cannot write to node with children\n" node.name;
                  (* false; *)
                  
                  (* XXX: HUGE HACK for broken tools *)
                  node.contents2 := Some (Value (d));
                  true;
                ) else (
                  node.contents := Value (d);
                  true;
                );
              );
        )
    | None -> true;
  );;

(* DO FUNCTIONS *)

(* Handle directory request *)
let do_directory conn in_buff =
  let name = canonicalise conn (strip_null !(in_buff.buffer)) in
  
  try (
      let node = get_node conn name [ XS_PERM_READ ] in
      
      match !(node.contents) with
      | Children (c) -> (
            let rec loop i lst =
              if (i < (List.length lst)) then (
                (null_terminate ((List.nth lst i).name)) ^ (loop (succ i) lst);
              ) else (
                String.create 0;
              );
            in
            
            let children = loop 0 c in
            
            send_reply conn XS_DIRECTORY children (String.length children);
          )
      | Value (v) -> send_reply conn XS_DIRECTORY "" 0;
    )
  with e -> (
        match e with
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Handle reading a node *)
let do_read conn in_buff =
  let name = canonicalise conn (strip_null !(in_buff.buffer)) in
  
  try (
      let node = get_node conn name [ XS_PERM_READ ] in
      
      let data =
        match !(node.contents) with
        | Children (c) -> (
              (* XXX: HUGE HACK for broken tools *)
              match !(node.contents2) with
              | Some (cont2) -> (
                    match (cont2) with
                    | Value (v) -> v
                    | Children (c) -> ""; (* XXX: This should never happen... *)
                  )
              | None ->"";
            )
        | Value (v) -> v
      in
      
      send_reply conn XS_READ data (String.length data);
    )
  with e -> (
        match e with
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Handle getting a node's permissions *)
let do_get_perms conn in_buff =
  let name = canonicalise conn (strip_null !(in_buff.buffer)) in
  
  try (
      let node = get_node conn name [ XS_PERM_READ ] in
      let data = perms_to_strings !(node.permissions) in
      
      send_reply conn XS_GET_PERMS data (String.length data);
    )
  with e -> (
        match e with
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Handle a watch *)
let do_watch conn in_buff =
  let null_index = String.index !(in_buff.buffer) (char_of_int 0) in
  let first_char = String.get !(in_buff.buffer) 0 in
  
  let relative =
    if (first_char = '@') then (
      false;
    ) else (
      not (first_char = '/');
    ) in
  
  let name =
    if (not (first_char = '@')) then (
      canonicalise conn (String.sub !(in_buff.buffer) 0 null_index);
    ) else (
      String.sub !(in_buff.buffer) 0 null_index;
    ) in
  
  if ((first_char = '@') && ((String.length name) > xenstore_rel_path_max)) then (
    send_error conn Unix.EINVAL;
  ) else (
    
    if ((first_char = '@') || (is_valid_nodename name)) then (
      let token = String.sub !(in_buff.buffer) (null_index + 1) ((String.length !(in_buff.buffer)) - (null_index + 1)) in
      
      (* Check for duplicates *)
      let dup_watch w = (
          (w.node = name) && (w.token = token);
        )
      in
      
      if (not (List.exists dup_watch !(conn.watches))) then (
        (* Check quota *)
        if ((get_domain_watches conn) < !quota_num_watches_per_domain) then (
          let watch = {
            events = ref [];
            
            relative_path =
              if (relative) then (
                get_implicit_path conn;
              ) else (
                None;
              );
            
            token = token;
            node = name;
          } in
          
          (* Increment the domain watch count *)
          domain_watches_inc conn;
          
          (* Add the watch to the connection *)
          conn.watches := List.append !(conn.watches) [ watch ];
          
          (* TODO: Tracing trace_create w "watch"; *)
          
          (* Send ACK to client *)
          send_ack conn XS_WATCH;
          
          (* Fire event once up front *)
          add_event conn watch watch.node;
        ) else (
          send_error conn Unix.E2BIG;
        );
      ) else (
        send_error conn Unix.EEXIST;
      );
    ) else (
      send_error conn Unix.EINVAL;
    );
  );;

(* Handle a unwatching a watch *)
let do_unwatch conn in_buff =
  let null_index = String.index !(in_buff.buffer) (char_of_int 0) in
  let first_char = String.get !(in_buff.buffer) 0 in
  
  let name =
    if (not (first_char = '@')) then (
      canonicalise conn (String.sub !(in_buff.buffer) 0 null_index);
    ) else (
      String.sub !(in_buff.buffer) 0 null_index;
    ) in
  
  let token = String.sub !(in_buff.buffer) (null_index + 1) ((String.length !(in_buff.buffer)) - (null_index + 1)) in
  
  let rec loop watches i = (
      if (i < (List.length watches)) then (
        let watch = List.nth watches i in
        
        if ((watch.node = name) && (watch.token = token)) then (
          (* Remove watch from list *)
          conn.watches := remove_from_list watches i;
          true;
        ) else (
          loop watches (succ i);
        );
      ) else (
        false;
      );
    )
  in
  
  if (loop !(conn.watches) 0) then (
    (* Decrease domain watch count *)
    domain_watches_dec conn;
    
    send_ack conn XS_UNWATCH;
  ) else (
    send_error conn Unix.ENOENT;
  );;

(* Handle the start of a transaction *)
let do_transaction_start conn in_buff =
  let f () = (
      let trans = {
        changes = ref [];
        changed_domains = ref [];
        
        generation = ref !generation;
        
        (* FIXME: Deep copy of the store *)
        store = ref (deep_node_copy !store);
        
        (* Pick an unused transaction identifier *)
        trans_id = ref (get_unused_transaction_id conn);
      
      } in
      
      conn.transaction_list := List.append !(conn.transaction_list) [ trans ];
      conn.transaction_started := Int64.succ !(conn.transaction_started);
      
      let data = (Int64.to_string !(trans.trans_id)) ^ (String.make 1 (char_of_int 0)) in
      
      send_reply conn XS_TRANSACTION_START data (String.length data);
    )
  in
  
  (* FIXME: We don't support nested transactions *)
  match !(conn.transaction) with
  | Some (trans) -> (
        if (trans = null_transaction) then (
          f ();
        ) else (
          send_error conn Unix.EBUSY;
        );
      )
  | None -> (
        (* Check if over quota *)
        if ((!(conn.id) > 0) && ((Int64.to_int !(conn.transaction_started)) > !quota_max_transaction)) then (
          send_error conn Unix.ENOSPC;
        ) else (
          f ();
        );
      );;

(* Handle the end of a transaction *)
let do_transaction_end conn in_buff =
  let data = strip_null !(in_buff.buffer) in
  
  if (((String.length data) = 0) || ((not (data = "T")) && (not (data = "F")))) then (
    send_error conn Unix.EINVAL;
  ) else (
    match !(conn.transaction) with
    | Some (trans) -> (
          conn.transaction := None;
          conn.transaction_started := Int64.pred !(conn.transaction_started);
          
          if (data = "T") then (
            (* FIXME: Merge, rather failing on any change *)
            if (!(trans.generation) = !generation) then (
              (* Replace store *)
              store := !(trans.store);
              (* trans.store := null_node; *)
              
              (* Fix domain entry for each changed domain *)
              let f dom = (
                  domain_entry_fix dom.changed_dom_id !(dom.changed_num_entries);
                )
              in
              
              List.iter f !(trans.changed_domains);
              
              (* Fire off the watches for everything that changed *)
              let fw node = (
                  fire_watches conn node.node_name node.recurse;
                ) in
              
              List.iter fw !(trans.changes);
              
              (* Increase transaction generation *)
              generation := Int64.succ !generation;
              
              send_ack conn XS_TRANSACTION_END;
            ) else (
              send_error conn Unix.EAGAIN;
            );
          ) else (
            send_ack conn XS_TRANSACTION_END;
          );
        )
    | None -> send_error conn Unix.ENOENT;
  );;

(* Handle introduce *)
let do_introduce conn in_buff =
  if (!(conn.id) = 0) then (
    let null_index = String.index !(in_buff.buffer) (char_of_int 0) in
    let null_index2 = String.index (String.sub !(in_buff.buffer) (null_index + 1) ((String.length !(in_buff.buffer)) - (null_index + 1))) (char_of_int 0) in
    let null_index3 = String.index (String.sub !(in_buff.buffer) (null_index2 + 1) ((String.length !(in_buff.buffer)) - (null_index2 + 1))) (char_of_int 0) in
    
    let dom_id = int_of_string (String.sub !(in_buff.buffer) 0 null_index) in
    let mfn = int_of_string (String.sub !(in_buff.buffer) (null_index + 1) null_index2) in
    let port = int_of_string (String.sub !(in_buff.buffer) (null_index + null_index2 + 2) null_index3) in
    
    (* Sanity check args *)
    if (port > 0) then (
      match find_domain_by_domid dom_id with
      | None -> (
            let interface = xc_map_foreign_range !xc_handle dom_id (get_page_size 0) mfn in
            
            match (new_domain dom_id (file_descr_of_int port)) with
            | Some (dom) -> (
                  (* Set final domain values *)
                  dom.interface := Some (interface);
                  dom.mfn := mfn;
                  
                  (* Fire watches *)
                  fire_watches null_conn "@introduceDomain" false;
                  
                  (* Reset domain *)
                  
                  (* Send ACK to client *)
                  send_ack conn XS_INTRODUCE;
                )
            | None -> (
                  (* unmap *)
                  ignore (munmap interface (get_page_size 0));
                  
                  (* TODO: get proper errno *)
                  send_error conn Unix.EACCES;
                );
          )
      | Some (dom) -> (
            if ((!(dom.mfn) = mfn) && (!(dom.conn) <> conn)) then (
              match !xce_handle with
              | Some (xce_h) -> (
                    if (int_of_file_descr !(dom.port) > 0) then (
                      ignore (xc_event_chan_unbind xce_h (Int32.of_int (int_of_file_descr !(dom.port))));
                    );
                    
                    let rc = Int32.to_int (xc_event_chan_bind_interdomain xce_h dom_id (Int32.of_int port)) in
                    
                    dom.port := (
                      if (rc = (- 1)) then (
                        file_descr_of_int 0;
                      ) else (
                        file_descr_of_int rc;
                      )
                    );
                    
                    dom.remote_port := file_descr_of_int port;
                  )
              | None -> send_error conn Unix.EINVAL;
            ) else (
              send_error conn Unix.EINVAL;
            );
          );
    ) else (
      send_error conn Unix.EINVAL;
    );
  ) else (
    send_error conn Unix.EACCES;
  );;

(* Release a domain *)
let do_release conn in_buff =
  if (!(conn.id) = 0) then (
    let dom_id = int_of_string (strip_null !(in_buff.buffer)) in
    
    match (find_domain_by_domid dom_id) with
    | Some (domain) -> (
          if (!(domain.conn) <> null_conn) then (
            domain.conn := null_conn;
            
            send_ack conn XS_RELEASE;
          ) else (
            send_error conn Unix.EINVAL;
          );
        )
    | None -> send_error conn Unix.ENOENT;
  ) else (
    send_error conn Unix.EACCES;
  );;

(* Return a domain's path *)
let do_get_domain_path conn in_buff =
  let dom_id_str = strip_null !(in_buff.buffer) in
  
  if ((String.length dom_id_str) > 0) then (
    (* TODO: This should be a constant or something... *)
    let path = null_terminate ("/local/domain/" ^ dom_id_str) in
    
    send_reply conn XS_GET_DOMAIN_PATH path (String.length path);
  ) else (
    send_error conn Unix.EINVAL;
  );;

let do_write conn in_buff =
  let null_index = String.index !(in_buff.buffer) (char_of_int 0) in
  let name = canonicalise conn (String.sub !(in_buff.buffer) 0 null_index) in
  let data = String.sub !(in_buff.buffer) (null_index + 1) ((String.length !(in_buff.buffer)) - (null_index + 1)) in
  
  try (
      let node = get_node conn name [ XS_PERM_WRITE ] in
      
      if (write_node conn node (Some (data)) !(node.permissions)) then (
        (* Add to transaction *)
        add_change_node !(conn.transaction) name false;
        
        (* Fire watches *)
        fire_watches conn name false;
        
        (* Send ACK to client *)
        send_ack conn XS_WRITE;
      ) else (
        send_error conn Unix.EINVAL;
      );
    )
  with e -> (
        match e with
        | Unix.Unix_error (Unix.ENOENT, _, _) -> (
              try (
                  ignore (
                      create_node conn name data;
                    );
                  
                  (* Add to transaction *)
                  add_change_node !(conn.transaction) name false;
                  
                  (* Fire watches *)
                  fire_watches conn name false;
                  
                  (* Send ACK to client *)
                  send_ack conn XS_WRITE;
                )
              with e2 -> (
                    match e2 with
                    | Unix.Unix_error (errno, _, _) -> send_error conn errno
                    | _ -> raise e2;
                  );
            )
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Create a new directory in the store *)
let do_mkdir conn in_buff =
  let name = canonicalise conn (strip_null !(in_buff.buffer)) in
  
  try (
      ignore(
          get_node conn name [ XS_PERM_WRITE ];
        );
      
      send_ack conn XS_MKDIR;
    )
  with e -> (
        match e with
        | Unix.Unix_error (Unix.ENOENT, _, _) -> (
              match (create_node conn name (String.create 0)) with
              | Some (node) -> (
                    (* Add to transaction *)
                    add_change_node !(conn.transaction) name false;
                    
                    (* Fire watches *)
                    fire_watches conn name false;
                    
                    (* Send ACK to client *)
                    send_ack conn XS_MKDIR;
                  )
              | None -> send_error conn Unix.ENOENT;
            )
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Remove a node from the store *)
let do_rm conn in_buff =
  let name = canonicalise conn (strip_null !(in_buff.buffer)) in
  
  try (
      ignore (
          get_node conn name [ XS_PERM_WRITE ];
        );
      
      if (name <> "/") then (
        (* Actually remove the node *)
        try (
            let parent_node = read_node conn (get_parent_name name) in
            
            (* Delete child *)
            if (delete_child conn parent_node name) then (
              (* XXX: Delete node? I think GC handles that... *)
              
              (* Add to transaction *)
              add_change_node !(conn.transaction) name true;
              
              (* Fire watches *)
              fire_watches conn name true;
              
              (* Send ACK to client *)
              send_ack conn XS_RM;
            ) else (
              send_error conn Unix.EINVAL;
            );
          )
        with e -> (
              match e with
              | Unix.Unix_error (_, _, _) -> send_error conn Unix.EINVAL
              | _ -> raise e;
            );
      ) else (
        send_error conn Unix.EINVAL;
      );
    )
  with e -> (
        match e with
        | Unix.Unix_error (Unix.ENOENT, _, _) -> (
              try (
                  ignore (
                      read_node conn (get_parent_name name);
                    );
                  
                  send_ack conn XS_RM;
                )
              with e -> (
                    match e with
                    | Unix.Unix_error (_, _, _) -> send_error conn Unix.ENOENT
                    | _ -> raise e;
                  );
            )
        | Unix.Unix_error (err, _, _) -> send_error conn err
        | _ -> raise e;
      );;

(* Set the permissions for a node *)
let do_set_perms conn in_buff =
  let null_index = String.index !(in_buff.buffer) (char_of_int 0) in
  let name = canonicalise conn (String.sub !(in_buff.buffer) 0 null_index) in
  let data = String.sub !(in_buff.buffer) (null_index + 1) ((String.length !(in_buff.buffer)) - (null_index + 1)) in
  
  (* Must own node to do this (tools can do this too) *)
  try (
      let node = get_node conn name [ XS_PERM_WRITE; XS_PERM_OWNER ] in
      let perms = xs_strings_to_perms data in
      
      (* Unprivileged domains may not change the owner *)
      if ((domain_is_unprivileged conn) && (!((List.hd perms).perm_dom_id) <> !((List.hd !(node.permissions)).perm_dom_id))) then (
        send_error conn Unix.EPERM;
      ) else (
        domain_entries_dec conn node;
        
        if (not (write_node conn node None perms)) then (
          send_error conn Unix.EACCES; (* XXX: errno *)
        ) else (
          domain_entries_inc conn node;
          
          add_change_node !(conn.transaction) name false;
          
          fire_watches conn name false;
          
          send_ack conn XS_SET_PERMS;
        );
      );
    )
  with e -> (
        match e with
        | Unix.Unix_error (errno, _, _) -> send_error conn errno
        | _ -> raise e;
      );;

(* Has a domain been introduced? *)
let do_is_domain_introduced conn in_buff =
  let dom_id = int_of_string (strip_null !(in_buff.buffer)) in
  
  let data = null_terminate (
        if (dom_id = dom_id_self) then (
          "T";
        ) else (
          match (find_domain_by_domid dom_id) with
          | Some (dom) -> "T"
          | None -> "F";
        )
      ) in
  
  send_reply conn XS_IS_DOMAIN_INTRODUCED data (String.length data);;

(* MESSAGE FUNCTIONS *)

(* Process a message *)
let process_message conn in_buff =
  (* XXX: Debug *)
  Printf.printf "process_message: conn id: %d msg_type: %s data: %s\n" !(conn.id) (sockmsg_string !(!(!(conn.in_buff).header).msg_type)) !(in_buff.buffer);
  flush stdout;
  
  (* TODO: Transaction stuff *)
  try (
      let trans = transaction_lookup conn !(!(in_buff.header).tx_id) in
      
      if (!(conn.transaction) = None) then (
        conn.transaction := trans;
        
        (* Check message type *)
        ignore (
            match !(!(in_buff.header).msg_type) with
            | XS_DEBUG -> send_error conn Unix.ENOSYS (* XXX: Going to be removed *)
            | XS_DIRECTORY -> do_directory conn in_buff
            | XS_READ -> do_read conn in_buff
            | XS_GET_PERMS -> do_get_perms conn in_buff
            | XS_WATCH -> do_watch conn in_buff
            | XS_UNWATCH -> do_unwatch conn in_buff
            | XS_TRANSACTION_START -> do_transaction_start conn in_buff
            | XS_TRANSACTION_END -> do_transaction_end conn in_buff
            | XS_INTRODUCE -> do_introduce conn in_buff
            | XS_RELEASE -> do_release conn in_buff
            | XS_GET_DOMAIN_PATH -> do_get_domain_path conn in_buff
            | XS_WRITE -> do_write conn in_buff
            | XS_MKDIR -> do_mkdir conn in_buff
            | XS_RM -> do_rm conn in_buff
            | XS_SET_PERMS -> do_set_perms conn in_buff
            | XS_IS_DOMAIN_INTRODUCED -> do_is_domain_introduced conn in_buff
            | XS_RESUME -> send_error conn Unix.ENOSYS (* XXX: Going to be removed *)
            | XS_SET_TARGET -> send_error conn Unix.ENOSYS (* XXX: Going to be removed *)
            | _ -> send_error conn Unix.ENOSYS; (* This should never happen *)
          );
        
        conn.transaction := None;
      ) else (
        Printf.printf "Connection already has transaction!\n";
        flush stdout;
        
        exit 1;
      );
    )
  with e -> (
        match e with
        | Not_found -> send_error conn Unix.ENOENT
        | _ -> raise e;
      );;

(* Consider a message *)
let consider_message conn =
  if (!verbose) then (
    Printf.printf "Got message %s len %i from %d\n" (sockmsg_string !(!(!(conn.in_buff).header).msg_type)) (Int64.to_int !(!(!(conn.in_buff).header).len)) !(conn.id);
    flush stdout;
  );
  
  process_message conn !(conn.in_buff);
  
  conn.in_buff := new_buffer ();;

(* HANDLE FUNCTIONS *)

(* Handle input from connection *)
let handle_input conn =
  let in_buff = !(conn.in_buff) in
  
  (* Not finished header yet? *)
  if !(in_buff.in_header) then (
    let bytes = (!(conn.read) conn (!(in_buff.header).buff) !(in_buff.used) (header_size - !(in_buff.used))) in
    
    if (bytes < 0) then (
      remove_connection conn;
      (* destroy_conn conn; *)
    ) else (
      in_buff.used := !(in_buff.used) + bytes;
      
      if (!(in_buff.used) = header_size) then (
        (* Populate header from buffer *)
        populate_header_variables !(in_buff.header) (!(in_buff.header).buff);
        
        (* TODO: Check PATH_MAX *)
        if ((Int64.to_int !(!(in_buff.header).len)) > xenstore_payload_max) then (
          match !syslog with
          | Some (log) -> Syslog.syslog log `LOG_ERR ("Client tried to feed us " ^ (Int64.to_string !(!(in_buff.header).len)))
          | None -> ();
              remove_connection conn;
          (* destroy_conn conn; *)
        ) else (
          
          in_buff.used := 0;
          in_buff.in_header := false;
          
          (* Populate data buffer *)
          in_buff.buffer := String.create (Int64.to_int !((!(in_buff.header)).len));
        );
      );
    );
  ) else (
    let bytes = (!(conn.read) conn !(in_buff.buffer) !(in_buff.used) ((Int64.to_int !(!(in_buff.header).len)) - !(in_buff.used))) in
    
    if (bytes < 0) then (
      remove_connection conn;
      (* destroy_conn conn; *)
    ) else (
      in_buff.used := !(in_buff.used) + bytes;
      
      if (!(in_buff.used) = (Int64.to_int !(!(in_buff.header).len))) then (
        trace_io conn "IN" in_buff;
        consider_message conn;
      );
    );
  );;

(* Handle output from connection *)
let handle_output conn =
  if (not (write_messages conn)) then (
    (* Free connection *)
    remove_connection conn;
    (* destroy_conn conn; *)
  );;

(* Handle event XXX: Domain function? *)
let handle_event () =
  let xce_handle_int =
    match !xce_handle with
    | Some (fd) -> int_of_file_descr fd
    | None -> (- 1);
  in
  
  let port_int = Int32.to_int (xc_event_chan_pending xce_handle_int) in
  
  if (port_int <> (- 1)) then (
    if (port_int = !virq_port) then (
      domain_cleanup ();
    );
    
    if ((Int32.to_int (xc_event_chan_unmask xce_handle_int (Int32.of_int port_int))) = (- 1)) then (
      barf_perror "Failed to write to event fd\n";
    );
  ) else (
    barf_perror "Failed to read from event fd\n";
  );;

(* MAIN FUNCTION *)

(* Main function *)
let main () =
  let do_fork = ref true in
  let output_pid = ref false in
  let no_domain_init = ref false in
  let event_chan_fd = ref None in
  let pid_file = ref "" in
  let (in_set : Unix.file_descr list ref) = ref [] in
  let (out_set : Unix.file_descr list ref) = ref [] in
  let timeout = ref 0.0 in
  
  (* Parse command-line arguments *)
  Arg.parse [
    ("--no-domain-init", Arg.Unit (fun () -> no_domain_init := true), "   to state that xenstored should not initialise dom0,");
    ("--pid-file", Arg.String (fun s -> pid_file := s), "<file>   giving a file for the daemon's pid to be written,");
    (* ("--help", Arg.Unit (Arg.Help), " to output this message,") *)
    ("--no-fork", Arg.Unit (fun () -> do_fork := false), "          to request that the daemon does not fork,");
    ("--output-pid", Arg.Unit (fun () -> output_pid := true), "       to request that the pid of the daemon is output,");
    ("--trace-file", Arg.String (fun s -> trace_file := s), "<file> giving the file for logging,");
    ("--entry-nb", Arg.Int (fun i -> quota_num_entry_per_domain := i), "<nb>     limit the number of entries per domain,");
    ("--entry-size", Arg.Int (fun i -> quota_max_entry_size := i), "<size> limit the size of entry per domain,");
    ("--entry-watch", Arg.Int (fun i -> quota_num_watches_per_domain := i),"<nb>  limit the number of watches per domain,");
    ("--transaction", Arg.Int (fun i -> quota_max_transaction := i), "<nb>  limit the number of transaction allowed per domain,");
    ("--no-recovery", Arg.Unit (fun () -> recovery := false), "      to request that no recovery should be attempted when\n                      the store is corrupted (debug only),");
    ("--preserve-local", Arg.Unit (fun () -> remove_local := false), "   to request that /local is preserved on start-up,");
    ("--verbose", Arg.Unit (fun () -> verbose := true), "          to request verbose execution.");
    ] parse_error usage;
  
  (* XXX: Debug *)
  Printf.printf "xenstored started, initialising...\n";
  flush stdout;
  
  (* Open log *)
  reopen_log ();
  
  (* Make sure xenstored directories exists *)
  xs_mkdir (xs_daemon_rundir ());
  xs_mkdir (xs_daemon_rootdir ());
  
  (* Fork *)
  if (!do_fork) then (
    syslog := Some (Syslog.openlog ~facility: (Syslog.facility_of_string "DAEMON") ~flags:[ `LOG_CONS ] "xenstored");
    daemonise ();
  );
  
  (* Write PID file *)
  if (!pid_file <> "") then (
    write_pid_file !pid_file;
  );
  
  (* Don't kill with SIGPIPE *)
  ignore (Sys.signal Sys.sigpipe Sys.Signal_ignore);
  
  (* Create sockets *)
  let socket = Unix.socket Unix.PF_UNIX Unix.SOCK_STREAM 0
  and socket_ro = Unix.socket Unix.PF_UNIX Unix.SOCK_STREAM 0 in
  
  (* Create addresses *)
  let addr = Unix.ADDR_UNIX (xs_daemon_socket ())
  and addr_ro = Unix.ADDR_UNIX (xs_daemon_socket_ro ()) in
  
  (* Remove old sockets if needed *)
  xs_unlink addr;
  xs_unlink addr_ro;
  
  (* Bind sockets *)
  Unix.bind socket addr;
  Unix.bind socket_ro addr_ro;
  
  (* Change permissions on sockets *)
  xs_chmod addr;
  xs_chmod addr_ro;
  
  (* Listen to sockets *)
  Unix.listen socket 1;
  Unix.listen socket_ro 1;
  
  (* Setup XenStore *)
  setup_store store;
  
  (* XXX: Debug *)
  print_store !store;
  
  (* Listen to hypervisor *)
  if (not !no_domain_init) then (
    ignore (domain_init ());
  );
  
  (* Restore existing connections *)
  restore_existing_connections ();
  
  (* Output the PID if required *)
  if (!output_pid) then (
    Printf.printf "%d\n" (Unix.getpid ());
    flush stdout;
  );
  
  (* If forking, redirect outputs to null device *)
  if (!do_fork) then (
    let dev_null = Unix.openfile "/dev/null" (Unix.O_RDWR :: []) 0x600 in
    
    Unix.dup2 dev_null Unix.stdin;
    Unix.dup2 dev_null Unix.stdout;
    Unix.dup2 dev_null Unix.stderr;
    Unix.close dev_null;
    (* TODO: xprintf := trace; *)
  );
  
  (* Set signal to reopen log *)
  ignore (Sys.signal Sys.sighup (Sys.Signal_handle trigger_reopen_log));
  
  (* Set event channel handle *)
  ignore (
      match !xce_handle with
      | Some (fd) -> event_chan_fd := Some (file_descr_of_int (Int32.to_int (xc_event_chan_fd fd)))
      | None -> ();
    );
  
  (* Get ready to listen to tools *)
  initialise_set in_set out_set socket socket_ro timeout;
  
  (* Tell kernel xenstored is up and running *)
  xenbus_notify_running ();
  
  (* XXX: Debug *)
  Printf.printf "xenstored initialised\n";
  flush stdout;
  
  (* Main loop *)
  let rec main_loop () = (
      let select_sets = Unix.select !in_set !out_set [] !timeout in
      
      (* TODO: reopen_log_pipe *)
      
      in_set := tuple_first select_sets;
      out_set := tuple_second select_sets;
      
      (* Check socket *)
      if (fd_is_set socket !in_set) then (
        accept_connection socket true;
      );
      
      (* Check read-only socket *)
      if (fd_is_set socket_ro !in_set) then (
        accept_connection socket_ro true;
      );
      
      (* Check event channel *)
      ignore (
          match !event_chan_fd with
          | Some (fd) -> (
                if (fd_is_set fd !in_set) then (
                  handle_event ();
                );
              )
          | None -> ();
        );
      
      (* Handle input and output for a connection *)
      let f conn = (
          ignore (
              match !(conn.domain) with
              | Some (d) -> (
                    if (domain_can_read conn) then (
                      handle_input conn;
                    );
                    
                    if ((domain_can_write conn) && (not ((List.length !(conn.out_list)) = 0))) then (
                      handle_output conn;
                    );
                  )
              | None -> (
                    ignore (
                        match !(conn.fd) with
                        | Some (fd) -> (
                              if (fd_is_set fd !in_set) then (
                                handle_input conn;
                              );
                              
                              if (fd_is_set fd !out_set) then (
                                handle_output conn;
                              );
                            )
                        | None -> ();
                      );
                  );
            );
        );
      in
      
      (* Handle input and output for connections *)
      List.iter f !connections;
      
      initialise_set in_set out_set socket socket_ro timeout;
      
      main_loop ();
    );
  in
  
  Printf.printf "starting main loop\n";
  ignore (main_loop ());;

(* Execute main function *)
main ();;
